from datetime import datetime
from utils_ak.builtin import update_dic


class Window:
    def __init__(self):
        self.state = "open"  # 'open', 'closed'

    def close(self):
        self.state = "closed"

    def is_closeable(self):
        # check if we can close the window now
        raise NotImplemented()


class ProcessingSessionWindow(Window):
    def __init__(self, gap):
        """
        :param gap: int, gap between sessions in seconds
        """
        super().__init__()
        self.values = []
        self.last_processing_time = None
        self.gap = gap

    def add(self, value):
        self.values.append(value)
        self.last_processing_time = datetime.now()

    def is_closeable(self):
        return (
            self.last_processing_time
            and (datetime.now() - self.last_processing_time).total_seconds() > self.gap
        )

    def close(self):
        self.state = "closed"
        return self.values


class CollectorWindow(Window):
    def __init__(self, fields):
        super().__init__()
        self.filled = {}  # {field: {symbol: value}}
        self.fields = fields

    def add(self, values):
        values = {
            k: v for k, v in values.items() if k in self.fields
        }  # remove extra fields
        self.filled = update_dic(self.filled, values)

    def is_closeable(self):
        return set(self.filled.keys()) == set(self.fields)

    def close(self):
        self.state = "closed"
        return self.filled


def test():
    collector = CollectorWindow(fields=["a", "b"])

    print(collector.add({"a": 1}))
    print(collector.is_closeable(), collector.state)
    print(collector.add({"b": 1}))
    print(collector.is_closeable(), collector.state)
    print(collector.close())
    print(collector.is_closeable(), collector.state)

    import time

    session = ProcessingSessionWindow(2)
    session.add("asdf")
    print(session.is_closeable())
    time.sleep(1)
    session.add("asdf")
    print(session.is_closeable())
    time.sleep(1)
    print(session.is_closeable())
    time.sleep(1)
    print(session.is_closeable())
    print(session.close())


if __name__ == "__main__":
    test()



""" Math functionality. """
from decimal import (
    ROUND_HALF_DOWN,
    ROUND_DOWN,
    ROUND_CEILING,
    ROUND_05UP,
    ROUND_FLOOR,
    ROUND_HALF_EVEN,
    ROUND_HALF_UP,
    ROUND_UP,
    Decimal as D,
)

ROUND_DIC = {
    "nearest_half_even": ROUND_HALF_EVEN,
    "floor": ROUND_FLOOR,
    "ceil": ROUND_CEILING,
    "down": ROUND_DOWN,
    "up": ROUND_UP,
    "nearest_half_up": ROUND_HALF_UP,
    "nearest_half_down": ROUND_HALF_DOWN,
    "down_except_0_and_5": ROUND_05UP,
}


def decimal_round(a, b, rounding="nearest_half_even", precision=0, strip=False):
    a, b = D(a), D(b)
    n = a / b / D(".1") ** precision
    rounding = ROUND_DIC.get(rounding, rounding)
    n = n.quantize(D("1."), rounding=rounding)
    res = n * b * D(".1") ** precision
    if strip:
        res = strip_decimal(res)
    return res


# be careful to use this for precise values: rounding 5. to 1. with floor method may result in 4 due to python floating system (5. == 4.99999999999) in python and stuff
def custom_round(
    a, b, rounding="nearest_half_even", precision=0, pre_round_precision=0
):
    if pre_round_precision:
        a = custom_round(a, b, "nearest_half_down", precision + pre_round_precision)
    n = a / b * (10 ** precision)
    n = int(decimal_round(n, 1, rounding))
    return n * b / (10 ** precision)


def cast_decimal(f, precision=15, strip=True):
    d = decimal_round(f, D("1.0"), precision=precision)
    if strip:
        s = str(d)
        s = strip_zeros(s)
        d = D(s)
    return d


def strip_decimal(d):
    return D(strip_zeros(str(d)))


def strip_zeros(s):
    return s.rstrip("0").rstrip(".") if "." in s else s


def test():
    roundings = [
        ROUND_HALF_DOWN,
        ROUND_DOWN,
        ROUND_CEILING,
        ROUND_05UP,
        ROUND_FLOOR,
        ROUND_HALF_EVEN,
        ROUND_HALF_UP,
        ROUND_UP,
    ]
    roundings += list(ROUND_DIC.keys())

    for rounding in roundings:
        for v in [1.99, 1.05, -1.99, -1.05]:
            print(rounding, v, 0.1, custom_round(v, 0.1, rounding))

    # 1.99
    print(decimal_round("1.99", "0.01", rounding="ROUND_FLOOR"))
    # may be 1.98 since 1.99 may be 1.9899999999
    print(decimal_round(1.99, "0.01", rounding="ROUND_FLOOR", precision=0))
    print(decimal_round(1.99, "0.01", rounding="ROUND_FLOOR", precision=2))
    # certainly 1.99, since we round 1.9899999 to 1.99 first
    print(decimal_round(1.99, "0.0100", rounding="ROUND_FLOOR", precision=10))
    print(decimal_round(1.99, "0.0100", rounding="ROUND_FLOOR", precision=10))
    print(
        decimal_round(1.99, "0.0100", rounding="ROUND_FLOOR", precision=10, strip=True)
    )

    print(custom_round(1.9999999, 1.0, rounding="floor"))
    print(custom_round(1.99999999, 1.0, rounding="floor", pre_round_precision=1))


if __name__ == "__main__":
    test()



import anyconfig
import copy
import os
import tempfile
from loguru import logger
from utils_ak.deployment.config import BASE_DIR
from utils_ak.deployment.controller import Controller
from utils_ak.coder import cast_js, cast_dict_or_list
from utils_ak.os import *
from utils_ak.dict import fill_template

from utils_ak.deployment.controller.test_controller import test_controller

# todo: success and failure responses (or errors?)


class DockerController(Controller):
    def start(self, deployment):
        id = deployment["id"]

        assert (
            len(deployment["containers"]) == 1
        ), "Only one-container pods are supported for now"

        entity, container = list(deployment["containers"].items())[0]
        params = {"entity": entity, "deployment_id": id, "image": container["image"]}

        config = cast_dict_or_list(
            os.path.join(BASE_DIR, "example/docker-compose.yml.template")
        )
        config = fill_template(config, **params)

        for k, v in container["command_line_arguments"].items():
            config["services"][entity]["command"].append(f"--{k}")
            config["services"][entity]["command"].append(cast_js(v))

        makedirs(f"data/docker-compose/{id}/")
        fn = f"data/docker-compose/{id}/docker-compose.yml"
        with open(fn, "w") as f:
            f.write(anyconfig.dumps(config, "yaml"))

        execute(f'docker-compose -f "{fn}" up -d --no-build')

    def stop(self, deployment_id):
        for id in self._get_docker_ids(deployment_id):
            execute(f"docker stop {id}")
            execute(f"docker rm {id}")
            # todo: review
            execute(f"docker network prune -f")  # clean networks
        remove_path(f"data/docker-compose/{deployment_id}/")

    def _get_docker_ids(self, deployment_id):
        ids = execute(f"docker ps -q -f name={deployment_id}*").split("\n")
        ids = [id for id in ids if id]
        return ids

    def log(self, deployment_id):
        # todo: log only deployment_id, not all
        ids = self._get_docker_ids(deployment_id)
        for id in ids:
            logger.debug("Logs for id", id=id, logs=execute(f"docker logs {id}"))


if __name__ == "__main__":
    test_controller(DockerController)



# mechanisms with async_tools
from multiprocessing.dummy import Pool as ThreadPool
from multiprocessing import Pool
from functools import wraps
import threading
from functools import partial
import time


# todo: implement other backends


def async_(f):
    @wraps(f)
    def inner(*args, **kwargs):
        thread = threading.Thread(target=f, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return inner


def run_async(f, *args, **kwargs):
    return async_(f)(*args, **kwargs)


def run_pool(callbacks, n_processes=4, Backend=ThreadPool):
    if len(callbacks) == 0:
        return []
    elif len(callbacks) == 1:
        # run in main process
        return callbacks[0]()
    else:
        n_processes = n_processes if len(callbacks) >= n_processes else len(callbacks)
        pool = Backend(n_processes)
        multiple_results = [pool.apply_async(f) for f in callbacks]
        res = [res.get() for res in multiple_results]
        pool.close()
        pool.join()
        return res


def run_map(f, args, n_processes=4, Backend=ThreadPool):
    return run_pool([partial(f, arg) for arg in args], n_processes, Backend)


def test():
    @async_
    def f(i):
        time.sleep(1)
        print(i)

    for i in range(10):
        f(i)

    print("Without join")

    threads = []
    for i in range(10):
        threads.append(f(i))
    for thread in threads:
        thread.join()

    print("With join")

    print("Run map")
    print(run_map(lambda x: x ** 2, [1, 2, 3]))

    print("Run pool")
    from functools import partial

    callbacks = [partial(lambda x: x ** 2, i) for i in [1, 2, 3]]
    print(run_pool(callbacks))

    def raise_exception(msg):
        raise Exception(msg)

    try:
        run_map(raise_exception, [1, 2, 3])
    except Exception as e:
        print(e)


if __name__ == "__main__":
    test()



import time
import zmq

from multiprocessing import Process

URL = "tcp://127.0.0.1:5556"
SYSTEM_URL = "tcp://127.0.0.1:5557"
POLL = True
# copy=True is much faster
COPY = True
SIZE = 1024
COUNT = 100_000
# lat or thr
now = time.perf_counter

from utils_ak.zmq.client import ZMQClient


# docs
# socket types: http://api.zeromq.org/4-1:zmq-socket
# some options dock. http://api.zeromq.org/2-1:zmq-setsockopt


def throughput_sink():
    context = zmq.Context()
    # router_system - server analogue
    router_system = context.socket(zmq.ROUTER)

    client = ZMQClient(context=context)
    client.subscribe(URL, conn_type="connect")

    router_system.bind(SYSTEM_URL)

    time.sleep(0.2)

    system_msg = router_system.recv_multipart()
    assert system_msg[1] == b"BEGIN", system_msg
    count = int(system_msg[2].decode("ascii"))
    router_system.send_multipart(system_msg)

    counter = 0
    while counter != count:
        msg = client.poll(0.01)
        if msg:
            counter += 1

    # system_msg[0] - instance to which we send message!
    router_system.send_multipart([system_msg[0], b"DONE"])

    router_system.close()
    context.term()


def get_throughput():
    context = zmq.Context()
    # dealer - client analogue
    system_dealer = context.socket(zmq.DEALER)
    client = ZMQClient()

    # no memory limits for sending
    system_dealer.SNDHWM = 0

    #  Add your socket options here.
    #  For example ZMQ_RATE, ZMQ_RECOVERY_IVL and ZMQ_MCAST_LOOP for PGM.

    system_dealer.connect(SYSTEM_URL)

    time.sleep(0.2)
    data = b" " * SIZE

    flags = 0
    system_dealer.send_multipart([b"BEGIN", str(COUNT).encode("ascii")])
    # Wait for the other side to connect.
    msg = system_dealer.recv_multipart()
    assert msg[0] == b"BEGIN"
    start = now()
    for i in range(COUNT):
        client.publish(URL, b"", data, conn_type="bind")
    sent = now()
    # wait for receiver
    reply = system_dealer.recv_multipart()
    elapsed = now() - start
    assert reply[0] == b"DONE"
    send_only = sent - start

    send_throughput = COUNT / send_only
    throughput = COUNT / elapsed
    megabits = throughput * SIZE * 8 / 1e6

    print("Throughput")
    print(f"Message size   : {SIZE}     [B]")
    print(f"Message count  : {COUNT}     [msgs]")
    print(f"Send only      : {send_throughput}     [msg/s]")
    print(f"Mean throughput: {throughput}     [msg/s]")
    print(f"Mean throughput: {megabits} [Mb/s]")
    print(f"Test time      : {elapsed} [s]")
    print()
    context.destroy()
    return (send_throughput, throughput)


def benchmark_throughput():
    p = Process(target=throughput_sink)
    p.start()
    get_throughput()
    p.join()


if __name__ == "__main__":
    benchmark_throughput()



import numpy as np


def is_float(obj):
    return np.issubdtype(type(obj), np.float)


def is_int(obj):
    return np.issubdtype(type(obj), np.integer)


def is_numeric(obj):
    return is_int(obj) or is_float(obj)


def is_int_like(obj):
    if isinstance(obj, str) or is_float(obj):
        try:
            obj = float(obj)
            return int(obj) == obj
        except:
            return False
    elif is_int(obj):
        return True
    else:
        return False


def test_is_int_like():
    print(is_int_like(1))
    print(is_int_like("1"))
    print(is_int_like("1.0"))
    print(is_int_like("1.2"))
    print(is_int_like("1.2;;"))



from .broker import Broker
from utils_ak.message_queue.clients.zmq_client import ZMQClient

ENDPOINT_TYPES = ["sub", "pub"]


class ZMQBroker(Broker):
    def __init__(self, endpoints):
        """
        :param endpoints: {collection_name: type, ...}
        """

        self.endpoints = endpoints

        # {endpoint: collection}
        self.collections = {}

        for collection, config in endpoints.items():
            if config["type"] in ["pub", "sub"]:
                self.collections[config["endpoint"]] = collection
            else:
                # forwarder
                self.collections[config["endpoint_pub"]] = collection
                self.collections[config["endpoint_sub"]] = collection

        self.cli = ZMQClient()

        self.async_supported = False

    def _get_endpoint_info(self, collection, action):
        """
        :param collection: str
        :param action: sub or pub
        :return:
        """
        config = self.endpoints[collection]
        endpoint_type = config["type"]
        if endpoint_type in ["pub", "sub"]:
            endpoint = config["endpoint"]
            conn_type = "bind" if endpoint_type == action else "connect"
        elif endpoint_type == "forwarder":
            if action == "pub":
                endpoint = config["endpoint_pub"]
            else:
                endpoint = config["endpoint_sub"]
            conn_type = "connect"
        else:
            raise Exception(f"Unsupported endpoint type {endpoint_type}")
        return endpoint, conn_type

    def subscribe(self, collection, topic):
        endpoint, conn_type = self._get_endpoint_info(collection, "sub")
        self.cli.subscribe(endpoint, topic, conn_type=conn_type)

    def publish(self, collection, topic, msg):
        endpoint, conn_type = self._get_endpoint_info(collection, "pub")
        self.cli.publish(endpoint, topic, msg, conn_type=conn_type)

    def poll(self, timeout=0.0):
        msg = self.cli.poll(timeout)
        if not msg:
            return
        endpoint, topic, msg = msg
        return self.collections[endpoint], topic, msg



import urllib
import json
import os
import re

from utils_ak.os import remove_path, execute


def get_current_notebook_path():
    from notebook import notebookapp
    import ipykernel

    """Returns the absolute path of the Notebook or None if it cannot be determined
    NOTE: works only when the security is token-based or there is also no password
    """
    connection_file = os.path.basename(ipykernel.get_connection_file())
    kernel_id = connection_file.split("-", 1)[1].split(".")[0]

    for srv in notebookapp.list_running_servers():
        try:
            if (
                srv["token"] == "" and not srv["password"]
            ):  # No token and no password, ahem...
                req = urllib.request.urlopen(srv["url"] + "api/sessions")
            else:
                req = urllib.request.urlopen(
                    srv["url"] + "api/sessions?token=" + srv["token"]
                )
            sessions = json.load(req)
            for sess in sessions:
                if sess["kernel"]["id"] == kernel_id:
                    return os.path.join(srv["notebook_dir"], sess["notebook"]["path"])
        except:
            pass  # There may be stale entries in the runtime directory
    return None


def get_current_notebook_code():
    path = get_current_notebook_path()
    path = path.replace("\\", "/")
    s = fr'jupyter nbconvert --to script "{path}"'
    execute(s)
    with open(path.replace(".ipynb", ".py"), "r") as f:
        code = f.read()
    remove_path(path.replace(".ipynb", ".py"))
    return code


def parse_cells(notebook_code):
    cell_headers = re.findall(r"(In\[[\d\s]+\])", notebook_code)
    for cell_header in cell_headers:
        notebook_code = re.sub(
            cell_header.replace("[", "\[").replace("]", "\]"), "In[]", notebook_code, 1
        )

    res = ""
    for line in notebook_code.split("\n"):
        if "In[]" in line:
            res += ">split_cell"
        else:
            res += line
            res += "\n"
    cells = res.split(">split_cell")

    return cells[1:]


def is_running_in_ipython():
    try:
        cfg = get_ipython().config
        if cfg["IPKernelApp"]["parent_appname"] == "ipython-notebook":
            return True
        else:
            return False
    except NameError:
        return False


def is_running_in_jupyter():
    # https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def safe_display(obj):
    if is_running_in_ipython():
        from IPython.display import display

        display(obj)
    else:
        print(obj)


def safe_markdown(obj):
    if is_running_in_ipython():
        from IPython.display import Markdown

        Markdown(obj)
    else:
        print(obj)



""" Python reflexion-level functionality. """
from functools import reduce
import importlib
import inspect
import sys


def extract_class(src, class_name):
    module = load(src)
    return reduce(getattr, class_name.split("."), module)


extract_class_by_module_name = extract_class


def extract_all_classes(src):
    module = load(src)
    return dict(
        [(name, cls) for name, cls in module.__dict__.items() if isinstance(cls, type)]
    )


def load(
    module_obj, module_name=None, reload=False, import_globals=False, globals_dic=None
):
    """
    :param import_globals: does not work for dependent modules properly. Use with ultimate care
    :return:
    """
    if inspect.ismodule(module_obj):
        module = module_obj
        if reload:
            module = importlib.reload(module)
    elif isinstance(module_obj, str):
        module_name = module_name or module_obj
        module = sys.modules.get(module_name)

        if not module or reload:
            spec = importlib.util.spec_from_file_location(module_name, module_obj)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
    else:
        raise Exception("Unknown module object")

    # todo: make better implementation
    # put all module variables into globals to reload
    globals_dic = globals_dic or globals()
    if import_globals:
        for x in dir(module):
            if not x.startswith("_"):
                globals_dic[x] = getattr(module, x)
    return module


cast_module = load



import anyconfig
import copy
import os
import tempfile

from loguru import logger

from utils_ak.deployment.config import BASE_DIR
from utils_ak.deployment.controller import Controller
from utils_ak.coder import cast_js, cast_dict_or_list
from utils_ak.os import *
from utils_ak.dict import fill_template


class KubernetesController(Controller):
    def start(self, deployment):
        id = deployment["id"]
        assert (
            len(deployment["containers"]) == 1
        ), "Only one-container pods are supported for now"

        # create docker-compose file and run it without building
        entity, container = list(deployment["containers"].items())[0]
        params = {"entity": entity, "deployment_id": id, "image": container["image"]}

        configs = cast_dict_or_list(
            os.path.join(BASE_DIR, "example/kubernetes.deployment.yml.template")
        )
        configs = [fill_template(config, **params) for config in configs]

        for config in configs:
            if config["kind"] == "ConfigMap":
                for k, v in container["command_line_arguments"].items():
                    config["data"][k.upper()] = cast_js(v)
        config_str = "---\n".join(
            [anyconfig.dumps(config, "yaml") for config in configs]
        )

        makedirs(f"data/kubernetes/{id}/")
        fn = f"data/kubernetes/{id}/kubernetes.yml"

        with open(fn, "w") as f:
            f.write(config_str)

        execute(f'kubectl apply -f "{fn}"')

    def stop(self, deployment_id):
        fn = f"data/kubernetes/{deployment_id}/kubernetes.yml"
        execute(f'kubectl delete -f "{fn}"')
        # remove_path(f'data/kubernetes/{deployment_id}/')

    def log(self, deployment_id):
        logger.debug(
            "Logs", logs=execute(f"kubectl logs -l deployment_id={deployment_id}")
        )


def test_kubernetes_controller():
    import anyconfig
    import time
    from utils_ak.loguru import logger, configure_loguru_stdout

    configure_loguru_stdout("DEBUG")

    deployment = anyconfig.load("../../example/deployment.yml")
    ctrl = KubernetesController()
    ctrl.start(deployment)
    time.sleep(10)
    ctrl.log(deployment["id"])
    ctrl.stop(deployment["id"])


if __name__ == "__main__":
    test_kubernetes_controller()



class Broker:
    async_supported = False

    def subscribe(self, collection, topic):
        raise NotImplementedError

    def publish(self, collection, topic, msg):
        raise NotImplementedError

    def poll(self, timeout=0.0):
        raise NotImplementedError

    async def aiopoll(self, timeout=0.0):
        raise NotImplementedError



import time
import multiprocessing

from mongoengine import connect as connect_to_mongodb

from utils_ak.simple_microservice import run_listener_async
from utils_ak.deployment import *
from utils_ak.loguru import configure_loguru_stdout

from utils_ak.job_orchestrator.job_orchestrator import JobOrchestrator
from utils_ak.job_orchestrator.models import *

from utils_ak.job_orchestrator.config import settings
from utils_ak.job_orchestrator.tests.test_monitor import run_monitor

MESSAGE_BROKER = settings.as_dict()["TRANSPORT"]["message_broker"]


def create_new_job(payload):
    configure_loguru_stdout("DEBUG")
    connect_to_mongodb(
        host=settings.job_queue.mongodb_host, db=settings.job_queue.mongodb_db
    )
    logger.info("Connected to mongodb")
    time.sleep(2)
    logger.debug("Creating new job...")
    Job.drop_collection()
    Worker.drop_collection()

    payload = dict(payload)
    payload.update(
        {
            "message_broker": MESSAGE_BROKER,
        }
    )

    job = Job(
        type="test",
        payload=payload,
        runnable={
            "image": "akadaner/test-worker",
            "python_main": "/Users/arsenijkadaner/Yandex.Disk.localized/master/code/git/python-utils-ak/utils_ak/job_orchestrator/worker/test/main.py",
        },
        running_timeout=15,
    )
    job.save()


def test_job_orchestrator(payload=None):
    configure_loguru_stdout("DEBUG")
    connect_to_mongodb(
        host=settings.job_queue.mongodb_host, db=settings.job_queue.mongodb_db
    )
    logger.info("Connected to mongodb")

    controller = ProcessController()

    run_listener_async("job_orchestrator", message_broker=MESSAGE_BROKER)
    job_orchestrator = JobOrchestrator(controller, MESSAGE_BROKER)
    multiprocessing.Process(target=run_monitor).start()
    if payload:
        multiprocessing.Process(target=create_new_job, args=(payload,)).start()
    job_orchestrator.run()


def test_success():
    test_job_orchestrator({"type": "batch"})


def test_stalled():
    test_job_orchestrator({"type": "batch", "stalled_timeout": 600})


def test_timeout():
    test_job_orchestrator({"type": "batch", "running_timeout": 20})


def test_failure():
    test_job_orchestrator({"type": "batch", "initializing_timeout": 600})


def test_run():
    test_job_orchestrator()


if __name__ == "__main__":
    # test_success()
    # test_stalled()
    # test_timeout()
    # test_failure()
    test_run()



"""
Formats:
    -   schemed str: scheme://ip:port or scheme://user:pass@user:pswd. Example: socks5://localhost:5556.
    -   dict: {'ip', 'port', 'scheme'} or {'ip', 'port', 'user', 'pswd', 'scheme'}. Scheme is 'http' by default.
    -   None
    Default format: dict.
"""

import re
from utils_ak import re as re_tools

# from here: https://stackoverflow.com/questions/106179/regular-expression-to-match-dns-hostname-or-ip-address
IP_PAT = "(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])"
HOST_PAT = "(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])"
PORT_PAT = "(\d+)"
SCHEME_PAT = "http|https|socks5"
USER_PAT = "[a-zA-Z0-9]+"
PSWD_PAT = "[^\s@]+"

PROXY_PAT = f"%SCHEME://(%USER:%PSWD@)?%HOST:%PORT"
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, "%SCHEME", "scheme", SCHEME_PAT)
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, "%USER", "user", USER_PAT)
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, "%PSWD", "pswd", PSWD_PAT)
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, "%HOST", "host", HOST_PAT)
PROXY_PAT = re_tools.replace_with_pattern(PROXY_PAT, "%PORT", "port", PORT_PAT)


def cast_dict(proxy_obj):
    if proxy_obj is None:
        return None
    elif isinstance(proxy_obj, dict):
        if any(x not in proxy_obj for x in ["scheme", "host", "port"]):
            raise ValueError("Missing some proxy keys")
        if ("user" in proxy_obj or "pswd" in proxy_obj) and any(
            x not in proxy_obj for x in ["user", "pswd"]
        ):
            raise ValueError("Missing some proxy keys")
        return proxy_obj
    elif isinstance(proxy_obj, str):
        search = re.search(PROXY_PAT, proxy_obj)
        if not search:
            raise ValueError(f"Non-valid proxy {proxy_obj}")
        return search.groupdict()
    else:
        raise ValueError("Unknown proxy type")


def cast_str(proxy_obj):
    proxy = cast_dict(proxy_obj)
    if not proxy:
        raise Exception("Cannot cast proxy string from null proxy")
    elif proxy.get("user"):
        scheme, user, pswd, host, port = (
            proxy["scheme"],
            proxy["user"],
            proxy["pswd"],
            proxy["host"],
            proxy["port"],
        )
        return f"{scheme}://{user}:{pswd}@{host}:{port}"
    else:
        scheme, host, port = proxy["scheme"], proxy["host"], proxy["port"]
        return f"{scheme}://{host}:{port}"


def cast_requests_proxies(proxy_obj):
    """
    :return:
       proxies = {
        'http': 'http://10.10.1.10:3128',
        'https': 'http://10.10.1.10:1080',
    }
    """
    if not proxy_obj:
        return None
    proxy_str = cast_str(proxy_obj)
    return {"http": proxy_str, "https": proxy_str}


if __name__ == "__main__":
    print(cast_dict("http://ip:1234"))
    print(cast_dict({"scheme": "http", "host": "host", "port": 9999}))
    print(
        cast_str(
            {
                "scheme": "http",
                "host": "host",
                "port": 9999,
                "user": "user",
                "pswd": "pswd",
            }
        )
    )
    print(cast_requests_proxies("http://user:pswd@host:9999"))



import fnmatch


class ScreenController:
    def run(self, filter=None):
        config = self.config
        df = self.list()
        if filter:
            config = {k: v for k, v in config.items() if fnmatch.fnmatch(k, filter)}
        for screen_name, path in config.items():
            if screen_name in df["name"].values:
                print(f"Already running: {screen_name}")
            else:
                print(f"Running: {screen_name}")
                self.run_python(screen_name, path)

    def cast_name(self, name):
        return f"{self.prefix}{name}"

    def state(self, pattern=None):
        config = self.config
        df = self.list()
        if pattern:
            config = {
                k: v for k, v in self.config.items() if fnmatch.fnmatch(k, pattern)
            }
        for screen_name, path in config.items():
            if screen_name not in df["name"].values:
                df.loc[len(df)] = [None, screen_name, None, "Not running"]
        return df.sort_values(by="name").reset_index(drop=True)



import logging

from utils_ak.clock import *
from utils_ak.dag import *
from utils_ak.coder import cast_js
from utils_ak.simple_event_manager import SimpleEventManager

from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors import pipe_connect, Stub


class FluidFlow:
    def __init__(self, root, verbose=False):
        self.root = root

        # create top if needed
        if len(self.root.leaves()) > 1:
            top = Stub("Top")
            for i, leaf in enumerate(self.root.leaves()):
                pipe_connect(leaf, top, f"Top parent {i}")

        self.verbose = verbose
        self.logger = logging.getLogger()

    def __str__(self):
        values = ["Flow:"]
        for node in self.root.iterate("down"):
            # values.append(' ' * 4 + str(node) + ': ' + cast_js(node.stats()))
            if node.display_stats():
                values.append(
                    " " * 4 + str(node) + ": " + cast_js(node.display_stats())
                )
        return "\n".join(values)

    def __repr__(self):
        return str(self)

    def log(self, *args):
        if self.verbose:
            # self.logger.debug(args)
            print(*args)

    def update(self, topic, ts, event):
        self.log("Processing time", ts)

        for method in [
            "update_value",
            "update_pressure",
            "update_speed",
            "update_triggers",
            "update_last_ts",
        ]:
            # self.log(f'Procedure {method}')
            for node in self.root.iterate("down"):
                getattr(node, method, lambda ts: None)(ts)
            # self.log(self)
        self.log(self)

        self.log()


def run_flow(flow):
    event_manager = SimpleEventManager()
    for node in flow.root.iterate("down"):
        node.set_event_manager(event_manager)
        node.subscribe()

    event_manager.subscribe("", flow.update)
    event_manager.add_event("update", 0, {})
    event_manager.run()



def to_lower_case(upper_case_name):
    res = ""
    for l in upper_case_name:
        if not res:
            if l.isupper():
                res += l.lower()
            else:
                res += l
        else:
            if l.isupper():
                res += "_"
                res += l.lower()
            else:
                res += l
    return res


def to_upper_case(lower_case_name, camel=False):
    values = lower_case_name.split("_")
    values = [v[0].upper() + v[1:] for v in values]
    if camel:
        first = values[0]
        values[0] = first[0].lower() + first[1:]
    return "".join(values)


def test():
    assert to_lower_case("camelCase") == "camel_case"
    assert to_lower_case("UpperCase") == "upper_case"
    assert to_lower_case("upper_case") == "upper_case"
    assert to_upper_case("upper_case") == "UpperCase"
    assert to_upper_case("upper_case", camel=True) == "upperCase"
    assert to_upper_case("UpperCase") == "UpperCase"
    assert to_upper_case("UpperCase", camel=True) == "upperCase"


if __name__ == "__main__":
    test()



import os
from uuid import uuid4

from utils_ak.architecture.integrity_guard.integrity_guard import IntegrityGuard
from utils_ak.os import remove_path, gen_tmp_fn

import logging

logger = logging.getLogger(__name__)


class FilesystemIntegrityGuard(IntegrityGuard):
    """ I'm an integrity guard. I check if some operation passed nicely or if it had crashed at some point. """

    def __init__(self, uuid=None):
        self.uuid = uuid or uuid4().hex

    @property
    def canary_path(self):
        return gen_tmp_fn(self.uuid, extension=".operation.canary")

    def is_integral(self):
        return not os.path.isfile(self.canary_path)

    def enter_unstable_state(self):  # Call me if you think that operation started
        logger.debug("Entering unstable state")
        open(self.canary_path, "w").close()  # Create a canary

    def leave_unstable_state(self):  # Call me if you think that operation ended
        logger.debug("Leaving unstable state")
        remove_path(self.canary_path)


if __name__ == "__main__":
    guard = FilesystemIntegrityGuard(uuid4().hex)
    guard.enter_unstable_state()  # Operation started, we are not integral!
    assert not guard.is_integral()
    guard.leave_unstable_state()  # We've finished nicely
    assert guard.is_integral()
